# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Callable, Dict, Optional, cast

import pyarrow as pa
from ray.data import Dataset

from deltacat.utils.url import DeltaCatUrl
from deltacat.io.datasink.deltacat_datasink import DeltaCatDatasink


class DeltaCatDataset(Dataset):
    @staticmethod
    def from_dataset(dataset: Dataset) -> DeltaCatDataset:
        # cast to DeltacatDataset in-place since it only adds new methods
        dataset.__class__ = DeltaCatDataset
        return cast(DeltaCatDataset, dataset)

    def write_deltacat(
        self,
        url: DeltaCatUrl,
        *,
        # if the source dataset only contains DeltaCAT metadata, then only copy the metadata to the destination... if it contains external source file paths, then register them in a new Delta.
        metadata_only: bool = False,
        # merge all deltas as part of the write operation
        copy_on_write: Optional[bool] = False,
        filesystem: Optional[pa.fs.S3FileSystem] = None,
        try_create_dir: bool = True,
        arrow_open_stream_args: Optional[Dict[str, Any]] = None,
        arrow_parquet_args_fn: Callable[[], Dict[str, Any]] = lambda: {},
        min_rows_per_file: Optional[int] = None,
        ray_remote_args: Dict[str, Any] = None,
        concurrency: Optional[int] = None,
        **arrow_parquet_args,
    ) -> None:
        """Writes the dataset to files and commits DeltaCAT metadata indexing
        the files written.

        This is only supported for datasets convertible to Arrow records.
        To control the number of files, use ``.repartition()``.

        Unless a custom block path provider is given, the format of the output
        files will be {uuid}_{block_idx}.{extension}, where ``uuid`` is a
        unique id for the dataset.

        The DeltaCAT manifest will be written to ``f"{path}/manifest``

        Examples:
            >>> ds.write_deltacat("s3://catalog/root/path")

        Time complexity: O(dataset size / parallelism)

        Args:
            url: The path to the root directory where materialized files and
                DeltaCAT manifest will be written.
            filesystem: The filesystem implementation to write to. This should
                be either a PyArrow S3FileSystem.
            try_create_dir: Try to create all directories in destination path
                if True. Does nothing if all directories already exist.
            arrow_open_stream_args: kwargs passed to
                pyarrow.fs.S3FileSystem.open_output_stream
            filename_provider: FilenameProvider implementation
                to write each dataset block to a custom output path.
            arrow_parquet_args_fn: Callable that returns a dictionary of write
                arguments to use when writing each block to a file. Overrides
                any duplicate keys from arrow_parquet_args. This should be used
                instead of arrow_parquet_args if any of your write arguments
                cannot be pickled, or if you'd like to lazily resolve the write
                arguments for each dataset block.
            arrow_parquet_args: Options to pass to
                pyarrow.parquet.write_table(), which is used to write out each
                block to a file.
        """
        datasink = DeltaCatDatasink(
            url,
            metadata_only=metadata_only,
            copy_on_write=copy_on_write,
            arrow_parquet_args_fn=arrow_parquet_args_fn,
            arrow_parquet_args=arrow_parquet_args,
            min_rows_per_file=min_rows_per_file,
            filesystem=filesystem,
            try_create_dir=try_create_dir,
            open_stream_args=arrow_open_stream_args,
            dataset_uuid=self._uuid,
        )
        self.write_datasink(
            datasink,
            ray_remote_args=ray_remote_args,
            concurrency=concurrency,
        )



from __future__ import annotations

from bisect import bisect_left, bisect_right
from collections import defaultdict
from dataclasses import dataclass
from itertools import tee
from typing import Any, Dict, Set, List, FrozenSet, Iterable, TypeVar, NamedTuple

from intervaltree import Interval, IntervalTree

from deltacat.experimental.storage.rivulet.metastore.delta import DeltaContext
from deltacat.experimental.storage.rivulet.metastore.sst import SSTable, SSTableRow
from deltacat.experimental.storage.rivulet import Schema

T = TypeVar("T")


# Can be replaced with itertools.pairwise once we're on python 3.10+
def pairwise(iterable):
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Block(NamedTuple):
    row: SSTableRow
    context: DeltaContext
    """Context from the manifest around the placement of this row in the LSM-Tree"""


@dataclass(frozen=True)
class BlockGroup:
    """
    TODO discuss "Block" terminology.

    This data structure represents a set of data blocks which have to be traversed
     together, because rows in the block is in an overlapping key range

    The data and key ranges within a BlockGroup can be processed in parallel during a scan

    Each block may show up in many BlockGroups. This is because as soon as the combination of blocks in the group
    change, it becomes a different group.

    Take the following example: A dataset has Field Group 1 (FG1) and Field Group 2 (FG2). The SSTs rows look like below:

    Field group 1     Field group 2
    ------------------|---------------
    [0,100] -> Block1 | [0-10]  -> Block4
    [3-90]  -> Block2 | [0-100] -> Block5
    [10-95] -> Block3 |

    If we want to scan this data, we can index of off the SST rows into an interval tree and observe the following
    "boundaries" in the interval tree:

    BlockGroup1 - Covers rows [0-3), includes Blocks 1,4,5
    BlockGroup2 - Covers rows [3-10), includes Blocks 1, 2, 4, 5
    BlockGroup3 - Covers rows [10-90), includes Blocks 1, 2, 3, 4*, 5
    BlockGroup4 - Covers rows [90-95), includes Blocks 1, 2*, 3, 5
    BlockGroup5 - Covers rows [95-100], includes Blocks 1, 3*, 5
    *special case - interval end==block group start

    Creating a sorted list of BlockGroups like this allows us to know exactly which blocks contain records
    for any given point query, range query, or scan. For instance, if the user queries for key=3, we know
    to read BlockGroup1, or key[0-10] to read BlockGroup1+BlockGroup2.
    """

    key_min: T
    """
    Key min is inclusive (the block group covers data where primary key>=key_min)
    """
    key_max: T
    field_group_to_blocks: Dict[Schema, FrozenSet[Block]]
    key_max_inclusive: bool = False
    """
    By default, key_min is inclusive and key_max is non-inclusive.
    For the highest key in an SSTable, we need to set key_max_inclusive to True
    """

    def key_in_range(self, key: T) -> bool:
        if self.key_max_inclusive:
            return self.key_min <= key <= self.key_max
        else:
            return self.key_min <= key < self.key_max

    def key_below_range(self, key: T) -> bool:
        return key < self.key_min


@dataclass(frozen=True)
class OrderedBlockGroups:
    """
    Ordered block groups representing a sequential interval of a primary key range

    Block groups have an inclusive

    The block groups have a "boundary table" which represents primary key lower/upper ranges in table

    For example, a boundary table of [1,3,5,10] has blocks: [1,3), [3,5), [5,10]
    """

    key_min: T
    """
    Key min is inclusive
    """
    key_max: T
    """
    Key max is inclusive
    """
    block_groups: List[BlockGroup]
    boundary_table: List[T]


class BlockIntervalTree:
    """
    This interval tree combines all L0 SSTables in order to effectively traverse key ranges
    so that data can be zippered across field groups

    This interval tree is SHARED across N different field groups. This is because
        maintaining a different interval tree for N different field groups would require
        more complex traversal to achieve the same result
    """

    def __init__(self):
        self.tree: IntervalTree = IntervalTree()
        self.max_key_map: Dict[Any, List[Interval]] = {}

    def add_sst_table(self, sst: SSTable, context: DeltaContext):
        """
        Add intervals to SSTree which use primary key min and max as intervals
        The data for each interval is a tuple of (schema, SSTableRow)
        """
        self.add_sst_rows(sst.rows, context)

    def add_sst_rows(self, sst_rows: Iterable[SSTableRow], context: DeltaContext):
        """
        Add individual SSTable rows to tree
        """
        for row in sst_rows:
            interval: Interval = Interval(row.key_min, row.key_max, Block(row, context))
            self.tree.add(interval)
            if row.key_max not in self.max_key_map:
                self.max_key_map[row.key_max] = [interval]
            else:
                self.max_key_map[row.key_max].append(interval)

    def get_sorted_block_groups(
        self, min_key: Any | None = None, max_key: Any | None = None
    ) -> OrderedBlockGroups:
        """
        Returns an ordered list of block group by primary key range
        The IntervalTree boundary table contains each boundary where the set of intervals change
        This function traverses the boundary table and builds a list

        Edge case - incompatibility with Interval Tree Library

        Note that the IntervalTree Library treats ALL ranges as min=inclusive, max=non-inclusive
        This is a different from our SSTables, where min_key and max_key are both inclusive.

        As a suboptimal workaround, we can fix this by adding an SSTRow to an interval when the
        SST row's max key is equal to the lower bound of the interval. Take the following example:

        Field group 1     Field group 2
        ------------------|---------------
        [0,100] -> Block1 | [0-10]  -> Block4
        [3-90]  -> Block2 | [0-100] -> Block5
        [10-95] -> Block3 |

        Our workaround adds for example Block4 to the interval [10,90), because Block4 is inclusive
        of key 10 but the IntervalTree library thinks it is non-inclusive.

        BlockGroup1 - Covers rows [0-3), includes Blocks 1,4,5
        BlockGroup2 - Covers rows [3-10), includes Blocks 1, 2, 4, 5
        BlockGroup3 - Covers rows [10-90), includes Blocks 1, 2, 3, 4*, 5
        BlockGroup4 - Covers rows [90-95), includes Blocks 1, 2*, 3, 5
        BlockGroup5 - Covers rows [95-100], includes Blocks 1, 3*, 5
        *special case - interval end==block group start

        An ideal solution would produce block groups like below. To do this against the IntervalTree
        library, we would need to know how to convert an inclusive range like ["bar", "foo"] into a
        range like ["bar", "fooa") where the end range is non-inclusive. It is fine for our block groups
        to be non-optimal, we just need code when zipper merging to detect if a block group's max key is
        less than the current iterator and therfore not consider it.

        Optimal block groups:
        BlockGroup1 - Covers rows [0-3), includes Blocks 1,5,4
        BlockGroup2 - Covers rows [3-10), includes Blocks 1,5,2,4
        BlockGroup3 - Covers rows [10-11), includes Blocks 1, 2, 3, 4, 5
        BlockGroup3 - Covers rows [11-91), includes Blocks 1,2,3,5
        BlockGroup4 - Covers rows [91-96), includes Blocks 1,3,5
        BlockGroup5 - Covers rows [96-100], includes Blocks 1,5

        :param: min_key optionally restrict result so that they must overlap or be greater than min_key, INCLUSIVE
            A range [0,200) will be included for min_key <=100 because it overlaps with the min key
            A range [100, 200) will be included because it overlaps with the min key (100 key inclusive)
        :param: max_key optionally restrict result to be less than or overlap with max_key, INCLUSIVE
            A range like [100,200) will NOT included for max key=200 because range maxes are non-inclusive
        """
        key_boundaries = self.tree.boundary_table.keys()
        block_groups: List[BlockGroup] = []
        block_groups_min = None
        block_groups_max = None

        # Note that we need to expand min_key_idx and max_key_idx by 1 to cover cases where
        # the pairwise traversal (x,y) has x>=min_key>=y and x<=max_key<=y
        if min_key is not None and max_key is not None and min_key > max_key:
            raise ValueError(
                f"min_key {min_key} cannot be greater than max_key {max_key}"
            )

        min_key_idx = (
            max(0, bisect_left(key_boundaries, min_key) - 1)
            if min_key is not None
            else None
        )
        max_key_idx = (
            bisect_right(key_boundaries, max_key) + 1 if max_key is not None else None
        )
        boundary_table = key_boundaries[min_key_idx:max_key_idx]

        for lower_bound, upper_bound in pairwise(boundary_table):
            # Note that IntervalTree library treats lower bound of slice as inclusive and upper as exclusive
            # We follow the same structure in our BlockGroup
            intervals: Set[Interval] = self.tree.overlap(lower_bound, upper_bound)

            # Special case for if max key is equal to lower_bound. See method pydoc for more details
            for i in self.max_key_map.get(lower_bound, []):
                intervals.add(i)

            field_group_to_blocks = defaultdict(set)
            for interval in intervals:
                data: Block = interval.data
                schema = data.context.schema
                field_group_to_blocks[schema].add(data)

            # freeze dict to make it hashable
            field_group_to_blocks = {
                k: frozenset(v) for k, v in field_group_to_blocks.items()
            }

            # Special case - if this is the very last iteration, set key_max_inclusive to True
            max_key_inclusive = upper_bound == boundary_table[-1]

            block_group = BlockGroup(
                lower_bound, upper_bound, field_group_to_blocks, max_key_inclusive
            )
            block_groups_min = (
                lower_bound
                if block_groups_min is None
                else min(block_groups_min, lower_bound)
            )
            block_groups_max = (
                upper_bound
                if block_groups_max is None
                else max(block_groups_max, upper_bound)
            )
            block_groups.append(block_group)

        return OrderedBlockGroups(
            block_groups_min, block_groups_max, block_groups, boundary_table
        )



import os
import logging
import deltacat as dc

from deltacat import logs
from deltacat import IcebergCatalog
from env import store_cli_args_in_os_environ

from pyiceberg.schema import (
    Schema,
    NestedField,
    DoubleType,
    StringType,
    TimestampType,
    FloatType,
    StructType,
)
from pyiceberg.partitioning import PartitionSpec, PartitionField
from pyiceberg.transforms import DayTransform, IdentityTransform
from pyiceberg.table.sorting import SortField, SortOrder

from deltacat.exceptions import TableAlreadyExistsError
from deltacat.experimental.storage.iceberg.model import (
    SchemaMapper,
    PartitionSchemeMapper,
    SortSchemeMapper,
)
from deltacat.env import create_ray_runtime_environment

# initialize the driver logger
driver_logger = logs.configure_application_logger(logging.getLogger(__name__))


def run(warehouse="s3://my-bucket/my/key/prefix", **kwargs):
    # create any runtime environment required to run the example
    runtime_env = create_ray_runtime_environment()

    # Start by initializing DeltaCAT and registering available Catalogs.
    # Ray will be initialized automatically via `ray.init()`.
    # Only the `iceberg` data catalog is provided so it will become the default.
    # If initializing multiple catalogs, use the `default_catalog_name` param
    # to specify which catalog should be the default.
    dc.init(
        catalogs={
            # the name of the DeltaCAT catalog is "iceberg"
            "iceberg": dc.Catalog(
                # Apache Iceberg implementation of deltacat.catalog.interface
                impl=IcebergCatalog,
                # kwargs for pyiceberg.catalog.load_catalog start here...
                # the name of the Iceberg catalog is "example-iceberg-catalog"
                name="example-iceberg-catalog",
                # for additional properties see:
                # https://py.iceberg.apache.org/configuration/
                properties={
                    "type": "glue",
                    "region_name": "us-east-1",
                    "warehouse": warehouse,
                },
            )
        },
        # pass the runtime environment into ray.init()
        ray_init_args={"runtime_env": runtime_env},
    )

    # define a native Iceberg table schema
    schema = Schema(
        NestedField(
            field_id=1, name="datetime", field_type=TimestampType(), required=True
        ),
        NestedField(field_id=2, name="symbol", field_type=StringType(), required=True),
        NestedField(field_id=3, name="bid", field_type=FloatType(), required=False),
        NestedField(field_id=4, name="ask", field_type=DoubleType(), required=False),
        NestedField(
            field_id=5,
            name="details",
            field_type=StructType(
                NestedField(
                    field_id=6,
                    name="created_by",
                    field_type=StringType(),
                    required=False,
                ),
            ),
            required=False,
        ),
    )

    # define a native Iceberg partition spec
    partition_spec = PartitionSpec(
        PartitionField(
            source_id=1, field_id=1000, transform=DayTransform(), name="datetime_day"
        )
    )

    # define a native Iceberg sort order
    sort_order = SortOrder(SortField(source_id=2, transform=IdentityTransform()))

    # create a table named `test_namespace.test_table`
    # we don't need to specify which catalog to create this table in since
    # only the "iceberg" catalog is available
    table_name = "test_table"
    namespace = "test_namespace"
    print(f"Creating Glue Table: {namespace}.{table_name}")
    try:
        table_definition = dc.create_table(
            table=table_name,
            namespace=namespace,
            schema=SchemaMapper.map(schema),
            partition_scheme=PartitionSchemeMapper.map(partition_spec, schema),
            sort_keys=SortSchemeMapper.map(sort_order, schema),
        )
        print(f"Created Glue Table: {table_definition}")
    except TableAlreadyExistsError:
        print(f"Glue Table `{namespace}.{table_name}` already exists.")

    print(f"Getting Glue Table: {namespace}.{table_name}")
    table_definition = dc.get_table(table_name, namespace)
    print(f"Retrieved Glue Table: {table_definition}")


if __name__ == "__main__":
    example_script_args = [
        (
            [
                "--warehouse",
            ],
            {
                "help": "S3 path for Iceberg file storage.",
                "type": str,
            },
        ),
        (
            [
                "--STAGE",
            ],
            {
                "help": "Example runtime environment stage (e.g. dev, alpha, beta, prod).",
                "type": str,
            },
        ),
    ]

    # store any CLI args in the runtime environment
    store_cli_args_in_os_environ(example_script_args)

    # run the example using os.environ as kwargs
    run(**os.environ)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, Optional, List

from deltacat.storage.model.metafile import Metafile
from deltacat.storage.model.locator import Locator, LocatorName

NamespaceProperties = Dict[str, Any]


class Namespace(Metafile):
    @staticmethod
    def of(
        locator: Optional[NamespaceLocator],
        properties: Optional[NamespaceProperties] = None,
    ) -> Namespace:
        namespace = Namespace()
        namespace.locator = locator
        namespace.properties = properties
        return namespace

    @property
    def locator(self) -> Optional[NamespaceLocator]:
        val: Dict[str, Any] = self.get("namespaceLocator")
        if val is not None and not isinstance(val, NamespaceLocator):
            self.locator = val = NamespaceLocator(val)
        return val

    @locator.setter
    def locator(self, namespace_locator: Optional[NamespaceLocator]) -> None:
        self["namespaceLocator"] = namespace_locator

    @property
    def namespace(self) -> Optional[str]:
        namespace_locator = self.locator
        if namespace_locator:
            return namespace_locator.namespace
        return None

    @property
    def properties(self) -> Optional[NamespaceProperties]:
        return self.get("properties")

    @properties.setter
    def properties(self, properties: Optional[NamespaceProperties]) -> None:
        self["properties"] = properties

    def url(self, catalog_name: Optional[str] = None) -> str:
        return (
            f"dc://{catalog_name}/{self.namespace}/"
            if catalog_name
            else f"namespace://{self.namespace}/"
        )


class NamespaceLocatorName(LocatorName):
    def __init__(self, locator: NamespaceLocator):
        self.locator = locator

    @property
    def immutable_id(self) -> Optional[str]:
        return None

    def parts(self) -> List[str]:
        return [self.locator.namespace]


class NamespaceLocator(Locator, dict):
    @staticmethod
    def of(namespace: Optional[str]) -> NamespaceLocator:
        namespace_locator = NamespaceLocator()
        namespace_locator.namespace = namespace
        return namespace_locator

    @property
    def name(self) -> NamespaceLocatorName:
        return NamespaceLocatorName(self)

    @property
    def parent(self) -> Optional[Locator]:
        return None

    @property
    def namespace(self) -> Optional[str]:
        return self.get("namespace")

    @namespace.setter
    def namespace(self, namespace: Optional[str]) -> None:
        self["namespace"] = namespace



# from deltacat.compute import index
import subprocess
import socket
import os
import time
import re

import deltacat as dc

from dataclasses import dataclass

from typing import Set, Optional, Dict, Any, Union

from ray.job_submission import JobSubmissionClient, JobStatus

from deltacat.utils.performance import timed_invocation


def _run_cmd(cmd: str) -> None:
    exit_code = int(os.system(cmd))
    assert exit_code == 0, f"`{cmd}` failed. Exit code: {exit_code}"


def _ray_up(
    cluster_cfg: str, cluster_name_override: str = None, restart_only: bool = False
) -> None:
    restart_flag = "--no-restart" if not restart_only else "--restart-only"
    cluster_name_option = (
        f"-n '{cluster_name_override}'" if cluster_name_override else ""
    )
    print(f"Starting Ray cluster from '{cluster_cfg}'")
    _run_cmd(
        f"ray up '{cluster_cfg}' -y --no-config-cache {restart_flag} {cluster_name_option} --disable-usage-stats"
    )
    print(f"Started Ray cluster from '{cluster_cfg}'")


def _is_port_in_use(port: Union[int, str]) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", int(port))) == 0


def _is_dashboard_running(port: Union[int, str]) -> bool:
    return _is_port_in_use(port)


def _ray_dashboard_up(
    cluster_cfg: str, port: Union[str, int], timeout_seconds=15
) -> None:
    print(f"Starting Ray Dashboard for Ray cluster '{cluster_cfg}'")
    _run_cmd(f"ray dashboard '{cluster_cfg}' --port {port} &")
    start = time.monotonic()
    dashboard_is_up = False
    while time.monotonic() - start <= timeout_seconds:
        if _is_dashboard_running(port):
            dashboard_is_up = True
            break
        time.sleep(0.1)
    if not dashboard_is_up:
        raise TimeoutError(
            f"Timed out after waiting {timeout_seconds} seconds for dashboard "
            f"to establish connection on port {port}."
        )
    print(f"Started Ray Dashboard for Ray cluster '{cluster_cfg}'")


def _get_head_node_ip(cluster_cfg: str) -> str:
    print(f"Getting Ray cluster head node IP for '{cluster_cfg}'")
    cmd = f"ray get-head-ip '{cluster_cfg}'"
    proc = subprocess.run(
        cmd,
        shell=True,
        capture_output=True,
        text=True,
        check=True,
    )
    # the head node IP should be the last line printed to stdout
    # TODO(pdames): add IPv6 support
    head_node_ip = proc.stdout.splitlines()[-1]
    if not re.match(
        r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$",
        head_node_ip,
    ):
        print(
            f"Failed to find Ray Head Node IP Address in `{cmd}` "
            f"output: {proc.stdout}"
        )
        raise RuntimeError("No Ray Head Node IP Address Found")
    print(f"Ray cluster head node IP for '{cluster_cfg}': {head_node_ip}")
    return head_node_ip


def _ray_down_cmd(cluster_cfg: str) -> str:
    return f"ray down '{cluster_cfg}' -y"


def _ray_down(cluster_cfg: str) -> None:
    print(f"Destroying Ray cluster for '{cluster_cfg}'")
    _run_cmd(_ray_down_cmd(cluster_cfg))
    print(f"Destroyed Ray cluster for '{cluster_cfg}'")


def _ray_cluster_running(cluster_cfg: str) -> bool:
    try:
        _get_head_node_ip(cluster_cfg)
    except Exception as e:
        print(f"Get Head Node IP Failed with Exception: {e}")
        print(f"Assuming Ray Cluster is Not Running")
        return False
    return True


@dataclass(frozen=True)
class DeltaCatJobRunResult:
    job_id: str
    job_status: JobStatus
    job_logs: Any


class DeltaCatJobClient(JobSubmissionClient):
    @staticmethod
    def of(
        cluster_cfg_file_path: str = "./deltacat.yaml",
        *,
        launch_cluster: bool = True,
        start_dashboard: bool = True,
        restart_ray: bool = False,
        head_node_ip: str = None,
        dashboard_wait_time_seconds: int = 30,
        port: Union[int, str] = "8265",
        cluster_name_override: str = None,
    ):
        job_submission_client_url = None
        try:
            # launch Ray cluster if necessary
            if cluster_cfg_file_path:
                if launch_cluster:
                    if not _ray_cluster_running(cluster_cfg_file_path) or restart_ray:
                        _ray_up(cluster_cfg_file_path, cluster_name_override)
                elif restart_ray:
                    if _ray_cluster_running(cluster_cfg_file_path):
                        _ray_up(
                            cluster_cfg_file_path, restart_ray, cluster_name_override
                        )
                    else:
                        raise RuntimeError(
                            f"Cannot Restart Ray: Ray Cluster for "
                            f"`{cluster_cfg_file_path}` not found."
                        )
                dashboard_running = _is_dashboard_running(port)
                if not dashboard_running and start_dashboard:
                    _ray_dashboard_up(
                        cluster_cfg=cluster_cfg_file_path,
                        port=port,
                        timeout_seconds=dashboard_wait_time_seconds,
                    )
                    dashboard_running = True
                if not head_node_ip:
                    head_node_ip = (
                        "127.0.0.1"
                        # use dashboard port forwarding on localhost
                        if dashboard_running
                        # fetch the remote head node IP
                        else _get_head_node_ip(cluster_cfg_file_path)
                    )
            else:
                head_node_ip = "127.0.0.1"
            job_submission_client_url = f"http://{head_node_ip}:{port}"
            print(
                f"Initializing Ray Job Submission Client with URL: "
                f"{job_submission_client_url}"
            )
            client = JobSubmissionClient(f"http://{head_node_ip}:{port}")
            # the below class change is safe as long as we only add new methods
            # to the wrapped JobSubmissionClient that don't alter its internal
            # state
            client.__class__ = DeltaCatJobClient
            return client
        except Exception as e:
            print(f"Unexpected error while initializing Ray Job Client: {e}")
            if job_submission_client_url:
                print(
                    f"Please ensure that Ray was installed with a job server "
                    f'enabled via `pip install -U "ray[default]"` and '
                    f"that http://{head_node_ip}:{port} is accessible. You "
                    f"can optionally run `ray dashboard` to forward the "
                    f"remote Ray head node port to a local port (default 8265) "
                    f'then run `ray_job_client("127.0.0.1", 8265)` '
                    f"to connect via localhost."
                )
            if cluster_cfg_file_path:
                print(
                    f"If you're done submitting jobs, ensure that the remote "
                    f"Ray Cluster is shut down by running: "
                    f"{_ray_down_cmd(cluster_cfg_file_path)}"
                )
            raise e

    def run_job(
        self,
        *,
        entrypoint: str,
        runtime_env: Optional[Dict[str, Any]] = None,
        timeout_seconds: int = 600,
        **kwargs,
    ) -> DeltaCatJobRunResult:
        """
        Synchronously submit and run a Ray job. This method combines Ray job submission and monitoring by submitting
        the job to the Ray Job Server, waiting for the job to complete,
        validating the job's terminal status, retrieving and returning job run
        result information if successful.

        Args:
            entrypoint: The entry point for the job to be executed (module
                or script to run)
            runtime_env: Runtime environment configuration for the job.
                Some commonly used keys include `working_dir` (directory
                containing the job code), `pip` (list of pip packages to
                install), and `env_vars` (environment variables for the job).
            timeout_seconds: Maximum time in seconds to wait for job completion.
                Default to 600 seconds (10 minutes).
            kwargs: Additional keyword arguments to pass to the job submission.

        Returns:
            Final results from the successful job run execution.

        Raises:
            RuntimeError: If the job fails or terminates with status other
                than SUCCEEDED.
            TimeoutError: If the job doesn't complete within the specified
                timeout period

        Example:
            >>> client = job_client()
            >>> logs = client.run_job(
            ...     # Shell command to run job
            ...     entrypoint="my_script.py",
            ...     runtime_env={
            ...         # Path to the local directory containing my_script.py
            ...         "working_dir": "./",
            ...         # Pip dependencies to install
            ...         "pip": ["pandas", "numpy"],
            ...         # System environment variables to set
            ...         "env_vars": {"DATA_PATH": "/path/to/data"},
            ...     },
            ...     timeout_seconds=1200
            ... )
        """

        job_id = self.submit_job(
            entrypoint=entrypoint,
            runtime_env=runtime_env,
            **kwargs,
        )
        job_status, latency = timed_invocation(
            self.await_job,
            job_id,
            timeout_seconds=timeout_seconds,
        )
        job_logs = self.get_job_logs(job_id)
        if job_status != JobStatus.SUCCEEDED:
            print(f"Job `{job_id}` logs: ")
            print(job_logs)
            raise RuntimeError(f"Job `{job_id}` terminated with status: {job_status}")
        return DeltaCatJobRunResult(
            job_id=job_id,
            job_status=job_status,
            job_logs=job_logs,
        )

    def await_job(
        self,
        job_id: str,
        await_status: Set[JobStatus] = {
            JobStatus.SUCCEEDED,
            JobStatus.STOPPED,
            JobStatus.FAILED,
        },
        *,
        timeout_seconds: int = 600,
    ) -> JobStatus:
        """
        Polls a job's status until it matches the desired status or times out.

        This function continuously checks the status of a specified job using the
        provided client. It will keep polling until either the desired status is
        reached or the timeout period expires.

        Args:
            job_id: The unique identifier of the job to monitor.
            await_status: Set of :class:`ray.job_submission.JobStatus` to wait for.
                The function will return when the job reaches any of these states.
            timeout_seconds: Maximum time to wait in seconds.
                Defaults to 600 seconds (10 minutes).

        Returns:
            The final status of the job.

        Raises:
            TimeoutError: If the desired status is not reached within the
            specified timeout period.

        Example:
            >>>
            >>> client = job_client()
            >>> job_id = client.submit_job(
            >>>     # Shell command to run job
            >>>     entrypoint=f"python copy.py --source '{source}' --dest '{dest}'",
            >>>     # Path to the local directory containing copy.py
            >>>     runtime_env={"working_dir": "./"},
            >>> )
            >>> # wait for the job to reach a terminal state
            >>> client.await_job(job_id)
        """
        start = time.monotonic()
        terminal_status = None
        while time.monotonic() - start <= timeout_seconds:
            status = self.get_job_status(job_id)
            if status in await_status:
                terminal_status = status
                break
            time.sleep(0.1)
        if not terminal_status:
            self.stop_job(job_id)
            raise TimeoutError(
                f"Timed out after waiting {timeout_seconds} seconds for job "
                f"`{job_id}` status: {status}"
            )
        return terminal_status


def local_job_client(*args, **kwargs) -> DeltaCatJobClient:
    """
    Create a DeltaCAT Job Client that can be used to submit jobs to a local Ray
    cluster. Initializes Ray if it's not already running.

    Args:
        *args: Positional arguments to pass to `deltacat.init()`.
        **kwargs: Keyword arguments to pass to `deltacat.init()`.
    Returns:
        DeltaCatJobClient: A client instance that can be used to submit and
            manage local Ray jobs.

    Raises:
        RuntimeError: If a local Ray Job Server cannot be found.
    """
    # force reinitialization to ensure that we can get the Ray context
    kwargs["force"] = True
    context = dc.init(*args, **kwargs)
    if context is None:
        raise RuntimeError("Failed to retrieve Ray context.")
    if context.dashboard_url:
        head_node_ip, port = context.dashboard_url.split(":")
    else:
        # the Ray Dashboard URL is also the Ray Job Server URL
        raise RuntimeError(
            "Ray Job Server not found! Please reinstall Ray using "
            "`pip install -U `ray[default]`"
        )
    return DeltaCatJobClient.of(
        None,
        launch_cluster=False,
        start_dashboard=False,
        head_node_ip=head_node_ip,
        port=port,
    )


def job_client(
    cluster_cfg_file_path: str = "./deltacat.yaml",
    *,
    launch_cluster: bool = True,
    start_dashboard: bool = True,
    restart_ray: bool = False,
    head_node_ip: str = None,
    dashboard_wait_time_seconds: int = 15,
    port: Union[str, int] = "8265",
    cluster_name_override: str = None,
) -> DeltaCatJobClient:
    """
    Create a DeltaCAT Job Client that can be used to submit jobs to a remote
    Ray cluster.

    Args:
        cluster_cfg_file_path: Path to the Ray Cluster Launcher
            Config file. Defaults to "./deltacat.yaml".
        launch_cluster : Whether to launch a new Ray cluster.
            Defaults to True.
        start_dashboard: Whether to start the Ray dashboard.
            Defaults to True.
        restart_ray: Whether to restart Ray if it's already
            running. Defaults to False.
        head_node_ip: IP address of the Ray cluster head node.
            If None, will use the configuration from the cluster config file.
            Defaults to None.
        dashboard_wait_time_seconds: Time in seconds to wait for the Ray
            dashboard to start if `start_dashboard` is True.
        port: Port number for the Ray
            dashboard/job server. Defaults to "8265".

    Returns:
        DeltaCatJobClient: A client instance that can be used to submit and
            manage jobs on the Ray cluster.

    Raises:
        RuntimeError: If the Ray Job Server is not found.
    """
    return DeltaCatJobClient.of(
        cluster_cfg_file_path,
        launch_cluster=launch_cluster,
        start_dashboard=start_dashboard,
        restart_ray=restart_ray,
        head_node_ip=head_node_ip,
        dashboard_wait_time_seconds=dashboard_wait_time_seconds,
        port=port,
        cluster_name_override=cluster_name_override,
    )



from unittest import TestCase
from deltacat.utils.pyarrow import (
    partial_parquet_file_to_table,
    pyarrow_read_csv,
    ContentTypeValidationError,
    content_type_to_reader_kwargs,
    _add_column_kwargs,
    file_to_table,
    file_to_parquet,
    table_to_file,
    ReadKwargsProviderPyArrowSchemaOverride,
    ReadKwargsProviderPyArrowCsvPureUtf8,
    RAISE_ON_DECIMAL_OVERFLOW,
    RAISE_ON_EMPTY_CSV_KWARG,
)
import decimal
from deltacat.types.media import ContentEncoding, ContentType
from deltacat.types.partial_download import PartialParquetParameters
from pyarrow.parquet import ParquetFile
import tempfile
import pyarrow as pa
from pyarrow import csv as pacsv
import fsspec
import gzip
import json
from pyarrow import (
    feather as paf,
    parquet as papq,
    orc as paorc,
)

PARQUET_FILE_PATH = "deltacat/tests/utils/data/test_file.parquet"
PARQUET_GZIP_COMPRESSED_FILE_PATH = "deltacat/tests/utils/data/test_file.parquet.gz"
EMPTY_UTSV_PATH = "deltacat/tests/utils/data/empty.csv"
NON_EMPTY_VALID_UTSV_PATH = "deltacat/tests/utils/data/non_empty_valid.csv"
OVERFLOWING_DECIMAL_PRECISION_UTSV_PATH = (
    "deltacat/tests/utils/data/overflowing_decimal_precision.csv"
)
OVERFLOWING_DECIMAL_SCALE_UTSV_PATH = (
    "deltacat/tests/utils/data/overflowing_decimal_scale.csv"
)
GZIP_COMPRESSED_FILE_UTSV_PATH = "deltacat/tests/utils/data/non_empty_compressed.gz"
BZ2_COMPRESSED_FILE_UTSV_PATH = "deltacat/tests/utils/data/non_empty_compressed.bz2"


class TestPartialParquetFileToTable(TestCase):
    def test_partial_parquet_file_to_table_sanity(self):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )

        self.assertEqual(
            partial_parquet_params.num_row_groups, 2, "test_file.parquet has changed."
        )

        # only first row group to be downloaded
        partial_parquet_params.row_groups_to_download.pop()

        result = partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            include_columns=["n_legs"],
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.columns), 1)

    def test_partial_parquet_file_to_table_when_schema_passed(self):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )
        # only first row group to be downloaded
        partial_parquet_params.row_groups_to_download.pop()

        schema = pa.schema(
            [
                pa.field("n_legs", pa.string()),
                pa.field("animal", pa.string()),
                # NOTE: This field is not in the parquet file, but will be added on as an all-null column
                pa.field("MISSING", pa.int64()),
            ]
        )

        pa_kwargs_provider = lambda content_type, kwargs: {"schema": schema}

        result = partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            pa_read_func_kwargs_provider=pa_kwargs_provider,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 3)

        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(0).name, "n_legs")
        self.assertEqual(result_schema.field(1).type, "string")
        self.assertEqual(result_schema.field(1).name, "animal")
        self.assertEqual(result_schema.field(2).type, "int64")
        self.assertEqual(result_schema.field(2).name, "MISSING")

    def test_partial_parquet_file_to_table_when_schema_missing_columns(self):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )
        # only first row group to be downloaded
        partial_parquet_params.row_groups_to_download.pop()

        schema = pa.schema(
            [
                pa.field("n_legs", pa.string()),
                pa.field("animal", pa.string()),
                # NOTE: This field is not in the parquet file, but will be added on as an all-null column
                pa.field("MISSING", pa.int64()),
            ]
        )

        pa_kwargs_provider = lambda content_type, kwargs: {"schema": schema}

        result = partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            pa_read_func_kwargs_provider=pa_kwargs_provider,
            partial_file_download_params=partial_parquet_params,
            column_names=["n_legs", "animal", "MISSING"],
            include_columns=["MISSING"],
        )

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 1)

        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "int64")
        self.assertEqual(result_schema.field(0).name, "MISSING")

    def test_partial_parquet_file_to_table_when_schema_passed_with_include_columns(
        self,
    ):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )
        # only first row group to be downloaded
        partial_parquet_params.row_groups_to_download.pop()

        schema = pa.schema(
            [pa.field("animal", pa.string()), pa.field("n_legs", pa.string())]
        )

        pa_kwargs_provider = lambda content_type, kwargs: {"schema": schema}

        result = partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            column_names=["n_legs", "animal"],
            pa_read_func_kwargs_provider=pa_kwargs_provider,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)

        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(0).name, "n_legs")  # order doesn't change

    def test_partial_parquet_file_to_table_when_multiple_row_groups(self):

        pq_file = ParquetFile(PARQUET_FILE_PATH)
        partial_parquet_params = PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        )

        self.assertEqual(
            partial_parquet_params.num_row_groups, 2, "test_file.parquet has changed."
        )

        result = partial_parquet_file_to_table(
            PARQUET_FILE_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            partial_file_download_params=partial_parquet_params,
        )

        self.assertEqual(len(result), 6)
        self.assertEqual(len(result.columns), 2)


class TestReadCSV(TestCase):
    def test_read_csv_sanity(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            None,
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        for index, field in enumerate(result_schema):
            self.assertEqual(field.name, schema.field(index).name)

        self.assertEqual(result.schema.field(0).type, "string")

    def test_read_csv_when_column_order_changes(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["ship_datetime_utc", "is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(1).type, "string")
        self.assertEqual(result_schema.field(0).type, "timestamp[us]")

    def test_read_csv_when_partial_columns_included(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result.column_names), 1)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")

    def test_read_csv_when_column_names_partial(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(ContentType.UNESCAPED_TSV.value, ["is_active"], None, kwargs)

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs),
        )

    def test_read_csv_when_excess_columns_included(self):

        schema = pa.schema(
            [
                ("is_active", pa.string()),
                ("ship_datetime_utc", pa.timestamp("us")),
                ("MISSING", pa.string()),
            ]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc", "MISSING"],
            ["is_active", "ship_datetime_utc", "MISSING"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(NON_EMPTY_VALID_UTSV_PATH, **kwargs),
        )

    def test_read_csv_when_empty_csv_sanity(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            None,
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)
        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)
        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(1).type, "timestamp[us]")

    def test_read_csv_when_empty_csv_include_columns(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["ship_datetime_utc", "is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(1).type, "string")
        self.assertEqual(result_schema.field(0).type, "timestamp[us]")

    def test_read_csv_when_empty_csv_include_partial_columns(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "ship_datetime_utc"],
            ["is_active"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 1)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")

    def test_read_csv_when_empty_csv_honors_column_names(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(ContentType.UNESCAPED_TSV.value, ["is_active"], None, kwargs)

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(EMPTY_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 0)
        self.assertEqual(len(result.column_names), 1)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")

    def test_read_csv_when_empty_csv_and_raise_on_empty_passed(self):

        schema = pa.schema(
            [("is_active", pa.string()), ("ship_datetime_utc", pa.timestamp("us"))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(ContentType.UNESCAPED_TSV.value, ["is_active"], None, kwargs)

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(
                EMPTY_UTSV_PATH, **{**kwargs, RAISE_ON_EMPTY_CSV_KWARG: True}
            ),
        )

    def test_read_csv_when_decimal_precision_overflows_and_raise_kwarg_specified(self):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal128(4, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )
        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)
        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(
                OVERFLOWING_DECIMAL_PRECISION_UTSV_PATH,
                **{**kwargs, RAISE_ON_DECIMAL_OVERFLOW: True},
            ),
        )

    def test_read_csv_when_decimal_precision_overflows_sanity(self):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal128(4, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(OVERFLOWING_DECIMAL_PRECISION_UTSV_PATH, **kwargs),
        )

    def test_read_csv_when_decimal_scale_overflows_and_raise_kwarg_specified(self):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal128(20, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )
        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(
                OVERFLOWING_DECIMAL_SCALE_UTSV_PATH,
                **{**kwargs, RAISE_ON_DECIMAL_OVERFLOW: True},
            ),
        )

    def test_read_csv_when_decimal_scale_overflows_sanity(self):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal128(20, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(OVERFLOWING_DECIMAL_SCALE_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(
            result[1][0].as_py(), decimal.Decimal("322236.66")
        )  # rounding decimal
        self.assertEqual(result[1][1].as_py(), decimal.Decimal("32.33"))  # not rounded
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(1).type, pa.decimal128(20, 2))

    def test_read_csv_when_decimal_scale_overflows_and_negative_scale(self):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal128(20, -2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(OVERFLOWING_DECIMAL_SCALE_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(
            result[1][0].as_py(),
            decimal.Decimal("322200"),  # consequence of negative scale
        )  # rounding decimal
        self.assertEqual(result[1][1].as_py(), decimal.Decimal("00"))
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(1).type, pa.decimal128(20, -2))

    def test_read_csv_when_decimal_scale_overflows_with_decimal256(self):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal256(20, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(OVERFLOWING_DECIMAL_SCALE_UTSV_PATH, **kwargs)

        self.assertEqual(len(result), 3)
        self.assertEqual(
            result[1][0].as_py(), decimal.Decimal("322236.66")
        )  # rounding decimal
        self.assertEqual(result[1][1].as_py(), decimal.Decimal("32.33"))  # not rounded
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(1).type, pa.decimal256(20, 2))

    def test_read_csv_when_decimal_scale_overflows_with_decimal256_and_raise_on_overflow(
        self,
    ):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal256(20, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowNotImplementedError,
            lambda: pyarrow_read_csv(
                OVERFLOWING_DECIMAL_SCALE_UTSV_PATH,
                **{**kwargs, RAISE_ON_DECIMAL_OVERFLOW: True},
            ),
        )

    def test_read_csv_when_decimal_scale_overflows_without_any_schema_then_infers(self):
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=None)
        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        result = pyarrow_read_csv(OVERFLOWING_DECIMAL_SCALE_UTSV_PATH, **kwargs)

        # The default behavior of pyarrow is to invalid skip rows
        self.assertEqual(len(result), 2)
        self.assertEqual(result[1][0].as_py(), 32.33)  # rounding decimal
        self.assertEqual(result[1][1].as_py(), 0.4)  # not rounded
        self.assertEqual(len(result.column_names), 2)
        result_schema = result.schema
        self.assertEqual(result_schema.field(0).type, "string")
        self.assertEqual(result_schema.field(1).type, pa.float64())

    def test_read_csv_when_decimal_scale_and_precision_overflow_and_raise_on_overflow(
        self,
    ):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal128(5, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        self.assertRaises(
            pa.lib.ArrowInvalid,
            lambda: pyarrow_read_csv(OVERFLOWING_DECIMAL_SCALE_UTSV_PATH, **kwargs),
        )

    def test_read_csv_when_decimal_scale_overflow_and_file_like_obj_passed(self):
        schema = pa.schema(
            [("is_active", pa.string()), ("decimal_value", pa.decimal128(15, 2))]
        )
        kwargs = content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value)
        _add_column_kwargs(
            ContentType.UNESCAPED_TSV.value,
            ["is_active", "decimal_value"],
            ["is_active", "decimal_value"],
            kwargs,
        )

        read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        kwargs = read_kwargs_provider(ContentType.UNESCAPED_TSV.value, kwargs)

        with open(OVERFLOWING_DECIMAL_SCALE_UTSV_PATH, "rb") as file:
            result = pyarrow_read_csv(file, **kwargs)

            self.assertEqual(len(result), 3)
            self.assertEqual(
                result[1][0].as_py(), decimal.Decimal("322236.66")
            )  # rounding decimal
            self.assertEqual(
                result[1][1].as_py(), decimal.Decimal("32.33")
            )  # not rounded
            self.assertEqual(len(result.column_names), 2)
            result_schema = result.schema
            self.assertEqual(result_schema.field(0).type, "string")
            self.assertEqual(result_schema.field(1).type, pa.decimal128(15, 2))


class TestWriters(TestCase):
    def setUp(self):
        self.table = pa.table({"col1": ["a,b\tc|d", "e,f\tg|h"], "col2": [1, 2]})
        self.fs = fsspec.filesystem("file")
        self.base_path = tempfile.mkdtemp()
        self.fs.makedirs(self.base_path, exist_ok=True)

    def tearDown(self):
        self.fs.rm(self.base_path, recursive=True)

    def test_write_feather(self):
        path = f"{self.base_path}/test.feather"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.FEATHER.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content
        result = paf.read_table(path)
        assert result.equals(self.table)

    def test_write_csv(self):
        path = f"{self.base_path}/test.csv.gz"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.CSV.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Should be quoted due to commas in data
                assert '"a,b\tc|d",1' in content
                assert '"e,f\tg|h",2' in content

    def test_write_tsv(self):
        path = f"{self.base_path}/test.tsv.gz"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.TSV.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Should be quoted due to tabs in data
                assert '"a,b\tc|d"\t1' in content
                assert '"e,f\tg|h"\t2' in content

    def test_write_psv(self):
        path = f"{self.base_path}/test.psv.gz"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PSV.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Should be quoted due to pipes in data
                assert '"a,b\tc|d"|1' in content
                assert '"e,f\tg|h"|2' in content

    def test_write_unescaped_tsv(self):
        # Create table without delimiters for unescaped TSV
        table = pa.table({"col1": ["abc", "def"], "col2": [1, 2]})
        path = f"{self.base_path}/test.tsv.gz"

        table_to_file(
            table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.UNESCAPED_TSV.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # With quoting_style="none", strings should not be quoted
                assert "abc\t1" in content
                assert "def\t2" in content

    def test_write_orc(self):
        path = f"{self.base_path}/test.orc"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.ORC.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content
        result = paorc.read_table(path)
        assert result.equals(self.table)

    def test_write_parquet(self):
        path = f"{self.base_path}/test.parquet"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content
        result = papq.read_table(path)
        assert result.equals(self.table)

    def test_write_json(self):
        path = f"{self.base_path}/test.json.gz"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.JSON.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content (should be GZIP compressed)
        with self.fs.open(path, "rb") as f:
            with gzip.GzipFile(fileobj=f) as gz:
                content = gz.read().decode("utf-8")
                # Each line should be a valid JSON object
                lines = [
                    line for line in content.split("\n") if line
                ]  # Skip empty lines
                assert len(lines) == 2  # 2 records
                assert json.loads(lines[0]) == {"col1": "a,b\tc|d", "col2": 1}
                assert json.loads(lines[1]) == {"col1": "e,f\tg|h", "col2": 2}

    def test_write_avro(self):
        import polars as pl

        path = f"{self.base_path}/test.avro"

        table_to_file(
            self.table,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.AVRO.value,
        )
        assert self.fs.exists(path), "file was not written"

        # Verify content by reading with polars
        result = pl.read_avro(path).to_arrow()
        # Cast the result to match the original table's schema
        # (the round-trip from arrow->polars->arrow casts string to large string)
        result = result.cast(self.table.schema)
        assert result.equals(self.table)


class TestPyArrowReaders(TestCase):
    def setUp(self):
        # Create test data files for reading
        self.fs = fsspec.filesystem("file")
        self.base_path = tempfile.mkdtemp()
        self.fs.makedirs(self.base_path, exist_ok=True)

        # Create test Table
        self.table = pa.Table.from_pylist(
            [
                {"col1": "a,b\tc|d", "col2": 1, "col3": 1.1},
                {"col1": "e,f\tg|h", "col2": 2, "col3": 2.2},
                {"col1": "test", "col2": 3, "col3": 3.3},
            ]
        )

        # Write test files in different formats
        self._create_test_files()

    def tearDown(self):
        self.fs.rm(self.base_path, recursive=True)

    def _create_test_files(self):
        # Create CSV file (GZIP compressed)
        csv_path = f"{self.base_path}/test.csv"
        with self.fs.open(csv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d",1,1.1\n"e,f\tg|h",2,2.2\ntest,3,3.3\n'
                gz.write(content.encode("utf-8"))

        # Create TSV file (GZIP compressed)
        tsv_path = f"{self.base_path}/test.tsv"
        with self.fs.open(tsv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d"\t1\t1.1\n"e,f\tg|h"\t2\t2.2\ntest\t3\t3.3\n'
                gz.write(content.encode("utf-8"))

        # Create PSV file (GZIP compressed)
        psv_path = f"{self.base_path}/test.psv"
        with self.fs.open(psv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d"|1|1.1\n"e,f\tg|h"|2|2.2\ntest|3|3.3\n'
                gz.write(content.encode("utf-8"))

        # Create unescaped TSV file (GZIP compressed)
        unescaped_tsv_path = f"{self.base_path}/test_unescaped.tsv"
        pa.Table.from_pylist(
            [
                {"col1": "abc", "col2": 1, "col3": 1.1},
                {"col1": "def", "col2": 2, "col3": 2.2},
                {"col1": "ghi", "col2": 3, "col3": 3.3},
            ]
        )
        with self.fs.open(unescaped_tsv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = "abc\t1\t1.1\ndef\t2\t2.2\nghi\t3\t3.3\n"
                gz.write(content.encode("utf-8"))

        # Create Parquet file
        parquet_path = f"{self.base_path}/test.parquet"
        with self.fs.open(parquet_path, "wb") as f:
            papq.write_table(self.table, f)

        # Create Feather file
        feather_path = f"{self.base_path}/test.feather"
        with self.fs.open(feather_path, "wb") as f:
            paf.write_feather(self.table, f)

        # Create JSON file (GZIP compressed)
        json_path = f"{self.base_path}/test.json"
        with self.fs.open(json_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                # Create NDJSON format - one JSON object per line
                lines = []
                for row in self.table.to_pylist():
                    lines.append(json.dumps(row))
                content = "\n".join(lines) + "\n"
                gz.write(content.encode("utf-8"))

        # Create Avro file using polars (since pyarrow delegates to polars for Avro)
        avro_path = f"{self.base_path}/test.avro"
        import polars as pl

        pl_df = pl.from_arrow(self.table)
        pl_df.write_avro(avro_path)

        # Create ORC file
        orc_path = f"{self.base_path}/test.orc"
        with self.fs.open(orc_path, "wb") as f:
            paorc.write_table(self.table, f)

    def test_content_type_to_reader_kwargs(self):
        # Test CSV kwargs
        csv_kwargs = content_type_to_reader_kwargs(ContentType.CSV.value)
        expected_csv = {"parse_options": pacsv.ParseOptions(delimiter=",")}
        assert (
            csv_kwargs["parse_options"].delimiter
            == expected_csv["parse_options"].delimiter
        )

        # Test TSV kwargs
        tsv_kwargs = content_type_to_reader_kwargs(ContentType.TSV.value)
        expected_tsv = {"parse_options": pacsv.ParseOptions(delimiter="\t")}
        assert (
            tsv_kwargs["parse_options"].delimiter
            == expected_tsv["parse_options"].delimiter
        )

        # Test PSV kwargs
        psv_kwargs = content_type_to_reader_kwargs(ContentType.PSV.value)
        expected_psv = {"parse_options": pacsv.ParseOptions(delimiter="|")}
        assert (
            psv_kwargs["parse_options"].delimiter
            == expected_psv["parse_options"].delimiter
        )

        # Test unescaped TSV kwargs
        unescaped_kwargs = content_type_to_reader_kwargs(
            ContentType.UNESCAPED_TSV.value
        )
        assert unescaped_kwargs["parse_options"].delimiter == "\t"
        assert unescaped_kwargs["parse_options"].quote_char is False
        assert unescaped_kwargs["convert_options"].null_values == [""]

        # Test Parquet kwargs (should be empty)
        parquet_kwargs = content_type_to_reader_kwargs(ContentType.PARQUET.value)
        assert parquet_kwargs == {}

        # Test ORC kwargs (should be empty)
        orc_kwargs = content_type_to_reader_kwargs(ContentType.ORC.value)
        assert orc_kwargs == {}

        # Test Avro kwargs (should be empty)
        avro_kwargs = content_type_to_reader_kwargs(ContentType.AVRO.value)
        assert avro_kwargs == {}

    def test_add_column_kwargs(self):
        kwargs = {}
        column_names = ["col1", "col2", "col3"]
        include_columns = ["col1", "col2"]

        # Test CSV column kwargs
        _add_column_kwargs(ContentType.CSV.value, column_names, include_columns, kwargs)
        assert kwargs["read_options"].column_names == column_names
        assert kwargs["convert_options"].include_columns == include_columns

        # Test Parquet column kwargs
        kwargs = {}
        _add_column_kwargs(
            ContentType.PARQUET.value, column_names, include_columns, kwargs
        )
        assert kwargs["columns"] == include_columns

    def test_file_to_table_csv(self):
        # Test reading CSV with file_to_table
        csv_path = f"{self.base_path}/test.csv"

        result = file_to_table(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.column("col1").to_pylist() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_table_tsv(self):
        # Test reading TSV with file_to_table
        tsv_path = f"{self.base_path}/test.tsv"

        result = file_to_table(
            tsv_path,
            ContentType.TSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.column("col1").to_pylist() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_table_psv(self):
        # Test reading PSV with file_to_table
        psv_path = f"{self.base_path}/test.psv"

        result = file_to_table(
            psv_path,
            ContentType.PSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.column("col1").to_pylist() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_table_unescaped_tsv(self):
        # Test reading unescaped TSV with file_to_table
        unescaped_tsv_path = f"{self.base_path}/test_unescaped.tsv"

        result = file_to_table(
            unescaped_tsv_path,
            ContentType.UNESCAPED_TSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.column("col1").to_pylist() == ["abc", "def", "ghi"]

    def test_file_to_table_parquet(self):
        # Test reading Parquet with file_to_table
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_table(
            parquet_path, ContentType.PARQUET.value, filesystem=self.fs
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.equals(self.table)

    def test_file_to_table_feather(self):
        # Test reading Feather with file_to_table
        feather_path = f"{self.base_path}/test.feather"

        result = file_to_table(
            feather_path, ContentType.FEATHER.value, filesystem=self.fs
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.equals(self.table)

    def test_file_to_table_json(self):
        # Test reading JSON with file_to_table
        json_path = f"{self.base_path}/test.json"

        result = file_to_table(
            json_path,
            ContentType.JSON.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
        )

        assert len(result) == 3
        assert set(result.column_names) == {"col1", "col2", "col3"}
        assert result.column("col1").to_pylist() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_table_avro(self):
        # Test reading Avro with file_to_table
        avro_path = f"{self.base_path}/test.avro"

        result = file_to_table(avro_path, ContentType.AVRO.value, filesystem=self.fs)

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        # Avro may have different dtypes, so compare values
        assert result.column("col1").to_pylist() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_table_orc(self):
        # Test reading ORC with file_to_table
        orc_path = f"{self.base_path}/test.orc"

        result = file_to_table(orc_path, ContentType.ORC.value, filesystem=self.fs)

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.equals(self.table)

    def test_file_to_table_with_column_selection(self):
        # Test reading with column selection
        csv_path = f"{self.base_path}/test.csv"

        result = file_to_table(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
            include_columns=["col1", "col2"],
        )

        assert len(result) == 3
        assert len(result.column_names) == 2  # Should only have 2 columns
        assert result.column_names == ["col1", "col2"]

    def test_file_to_table_with_kwargs_provider(self):
        # Test reading with kwargs provider
        csv_path = f"{self.base_path}/test.csv"
        provider = ReadKwargsProviderPyArrowCsvPureUtf8(
            include_columns=["col1", "col2", "col3"]
        )

        result = file_to_table(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
            pa_read_func_kwargs_provider=provider,
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        # With string types provider, all columns should be strings
        for col_name in result.column_names:
            assert result.schema.field(col_name).type == pa.string()

    def test_file_to_table_filesystem_inference(self):
        # Test filesystem inference when no filesystem is provided
        # Use JSON file since it should work well with inference
        json_path = f"{self.base_path}/test.json"

        result = file_to_table(
            json_path,
            ContentType.JSON.value,
            ContentEncoding.GZIP.value
            # No filesystem provided - should be inferred
        )

        assert len(result) == 3
        assert set(result.column_names) == {"col1", "col2", "col3"}
        assert result.column("col1").to_pylist() == ["a,b\tc|d", "e,f\tg|h", "test"]

    def test_file_to_table_unsupported_content_type(self):
        # Test error handling for unsupported content type
        parquet_path = f"{self.base_path}/test.parquet"

        with self.assertRaises(NotImplementedError) as context:
            file_to_table(parquet_path, "unsupported/content-type", filesystem=self.fs)

        assert "not implemented" in str(context.exception)

    def test_file_to_table_bzip2_compression(self):
        # Test BZIP2 compression handling
        import bz2

        # Create a BZIP2 compressed CSV file
        csv_content = '"a,b\tc|d",1,1.1\n"e,f\tg|h",2,2.2\ntest,3,3.3\n'
        compressed_content = bz2.compress(csv_content.encode("utf-8"))

        bz2_path = f"{self.base_path}/test.csv.bz2"
        with self.fs.open(bz2_path, "wb") as f:
            f.write(compressed_content)

        result = file_to_table(
            bz2_path,
            ContentType.CSV.value,
            ContentEncoding.BZIP2.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert len(result) == 3
        assert result.column_names == ["col1", "col2", "col3"]
        assert result.column("col1").to_pylist() == ["a,b\tc|d", "e,f\tg|h", "test"]


class TestFileToParquet(TestCase):
    def setUp(self):
        # Create test data files for reading
        self.fs = fsspec.filesystem("file")
        self.base_path = tempfile.mkdtemp()
        self.fs.makedirs(self.base_path, exist_ok=True)

        # Create test Table
        self.table = pa.Table.from_pylist(
            [
                {"col1": "a,b\tc|d", "col2": 1, "col3": 1.1},
                {"col1": "e,f\tg|h", "col2": 2, "col3": 2.2},
                {"col1": "test", "col2": 3, "col3": 3.3},
            ]
        )

        # Write test parquet files
        self._create_test_files()

    def tearDown(self):
        self.fs.rm(self.base_path, recursive=True)

    def _create_test_files(self):
        # Create basic Parquet file
        parquet_path = f"{self.base_path}/test.parquet"
        with self.fs.open(parquet_path, "wb") as f:
            papq.write_table(self.table, f)

        # Create larger Parquet file with multiple row groups
        large_table = pa.Table.from_pylist(
            [{"col1": f"row_{i}", "col2": i, "col3": float(i)} for i in range(1000)]
        )
        large_parquet_path = f"{self.base_path}/test_large.parquet"
        with self.fs.open(large_parquet_path, "wb") as f:
            papq.write_table(
                large_table, f, row_group_size=100
            )  # Create multiple row groups

    def test_file_to_parquet_basic(self):
        # Test basic parquet file reading
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_parquet(parquet_path, filesystem=self.fs)

        assert isinstance(result, papq.ParquetFile)
        assert result.num_row_groups > 0
        assert result.metadata.num_rows == 3
        assert result.metadata.num_columns == 3

        # Verify we can read the data
        table = result.read()
        assert len(table) == 3
        assert table.column_names == ["col1", "col2", "col3"]

    def test_file_to_parquet_with_schema_provider(self):
        # Test with schema override provider
        parquet_path = f"{self.base_path}/test.parquet"

        schema = pa.schema(
            [
                pa.field("col1", pa.string()),
                pa.field("col2", pa.string()),  # Override to string
                pa.field("col3", pa.string()),  # Override to string
            ]
        )

        provider = ReadKwargsProviderPyArrowSchemaOverride(schema=schema)

        result = file_to_parquet(
            parquet_path, filesystem=self.fs, pa_read_func_kwargs_provider=provider
        )

        assert isinstance(result, papq.ParquetFile)
        # Note: schema override might not affect ParquetFile metadata,
        # but should work when reading the table
        table = result.read()
        assert len(table) == 3

    def test_file_to_parquet_with_custom_kwargs(self):
        # Test with custom ParquetFile kwargs
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_parquet(
            parquet_path,
            filesystem=self.fs,
            validate_schema=True,  # Custom kwarg for ParquetFile
            memory_map=True,  # Another custom kwarg
        )

        assert isinstance(result, papq.ParquetFile)
        assert result.metadata.num_rows == 3

    def test_file_to_parquet_filesystem_inference(self):
        # Test filesystem inference when no filesystem is provided
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_parquet(
            parquet_path
            # No filesystem provided - should be inferred
        )

        assert isinstance(result, papq.ParquetFile)
        assert result.metadata.num_rows == 3
        assert result.metadata.num_columns == 3

    def test_file_to_parquet_large_file(self):
        # Test with larger parquet file (multiple row groups)
        large_parquet_path = f"{self.base_path}/test_large.parquet"

        result = file_to_parquet(large_parquet_path, filesystem=self.fs)

        assert isinstance(result, papq.ParquetFile)
        assert result.metadata.num_rows == 1000
        assert result.num_row_groups > 1  # Should have multiple row groups

        # Test reading specific row groups
        first_row_group = result.read_row_group(0)
        assert len(first_row_group) <= 100  # Based on row_group_size=100

    def test_file_to_parquet_metadata_access(self):
        # Test accessing various metadata properties
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_parquet(parquet_path, filesystem=self.fs)

        # Test metadata access
        metadata = result.metadata
        assert metadata.num_rows == 3
        assert metadata.num_columns == 3
        assert metadata.num_row_groups >= 1

        # Test schema access
        schema = result.schema
        assert len(schema) == 3
        assert "col1" in schema.names
        assert "col2" in schema.names
        assert "col3" in schema.names

        # Test schema_arrow property
        schema_arrow = result.schema_arrow
        assert isinstance(schema_arrow, pa.Schema)
        assert len(schema_arrow) == 3

    def test_file_to_parquet_column_selection(self):
        # Test reading specific columns
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_parquet(parquet_path, filesystem=self.fs)

        # Read only specific columns
        table = result.read(columns=["col1", "col2"])
        assert len(table.column_names) == 2
        assert table.column_names == ["col1", "col2"]
        assert len(table) == 3

    def test_file_to_parquet_invalid_content_type(self):
        # Test error handling for invalid content type
        parquet_path = f"{self.base_path}/test.parquet"

        with self.assertRaises(ContentTypeValidationError) as context:
            file_to_parquet(
                parquet_path,
                content_type=ContentType.CSV.value,  # Invalid content type
                filesystem=self.fs,
            )

        assert "cannot be read into pyarrow.parquet.ParquetFile" in str(
            context.exception
        )

    def test_file_to_parquet_invalid_content_encoding(self):
        # Test error handling for invalid content encoding
        parquet_path = f"{self.base_path}/test.parquet"

        with self.assertRaises(ContentTypeValidationError) as context:
            file_to_parquet(
                parquet_path,
                content_encoding=ContentEncoding.GZIP.value,  # Invalid encoding
                filesystem=self.fs,
            )

        assert "cannot be read into pyarrow.parquet.ParquetFile" in str(
            context.exception
        )

    def test_file_to_parquet_different_filesystems(self):
        # Test with different filesystem implementations
        parquet_path = f"{self.base_path}/test.parquet"

        # Test with fsspec filesystem
        result_fsspec = file_to_parquet(parquet_path, filesystem=self.fs)
        assert isinstance(result_fsspec, papq.ParquetFile)
        assert result_fsspec.metadata.num_rows == 3

        # Test with None filesystem (inferred)
        result_inferred = file_to_parquet(parquet_path, filesystem=None)
        assert isinstance(result_inferred, papq.ParquetFile)
        assert result_inferred.metadata.num_rows == 3

    def test_file_to_parquet_lazy_loading(self):
        # Test that ParquetFile provides lazy loading capabilities
        large_parquet_path = f"{self.base_path}/test_large.parquet"

        result = file_to_parquet(large_parquet_path, filesystem=self.fs)

        # ParquetFile should be created without loading all data
        assert isinstance(result, papq.ParquetFile)
        assert result.metadata.num_rows == 1000

        # Test reading only specific columns (lazy loading)
        partial_table = result.read(columns=["col1", "col2"])
        assert len(partial_table) == 1000  # All rows but only 2 columns
        assert partial_table.column_names == ["col1", "col2"]

        # Test reading specific row group (lazy loading)
        row_group_table = result.read_row_group(0)
        assert len(row_group_table) <= 100  # Based on row_group_size

    def test_file_to_parquet_performance_timing(self):
        # Test that performance timing is logged (basic functionality test)
        parquet_path = f"{self.base_path}/test.parquet"

        # This should complete without error and log timing
        result = file_to_parquet(parquet_path, filesystem=self.fs)

        assert isinstance(result, papq.ParquetFile)
        assert result.metadata.num_rows == 3


class TestFileToTableFilesystems(TestCase):
    """Test file_to_table with different filesystem implementations across all content types."""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self._create_test_files()

    def tearDown(self):
        import shutil

        shutil.rmtree(self.tmpdir)

    def _create_test_files(self):
        """Create test files for all supported content types."""
        # Test data
        test_data = pa.table(
            {
                "id": [1, 2, 3, 4, 5],
                "name": ["Alice", "Bob", "Charlie", "Diana", "Eve"],
                "age": [25, 30, 35, 28, 32],
                "score": [85.5, 92.0, 78.5, 88.0, 95.5],
            }
        )

        # File paths
        self.csv_file = f"{self.tmpdir}/test.csv"
        self.tsv_file = f"{self.tmpdir}/test.tsv"
        self.psv_file = f"{self.tmpdir}/test.psv"
        self.unescaped_tsv_file = f"{self.tmpdir}/test_unescaped.tsv"
        self.parquet_file = f"{self.tmpdir}/test.parquet"
        self.feather_file = f"{self.tmpdir}/test.feather"
        self.json_file = f"{self.tmpdir}/test.json"
        self.orc_file = f"{self.tmpdir}/test.orc"
        self.avro_file = f"{self.tmpdir}/test.avro"

        # Create CSV file
        pacsv.write_csv(
            test_data,
            self.csv_file,
            write_options=pacsv.WriteOptions(delimiter=",", include_header=False),
        )

        # Create TSV file
        pacsv.write_csv(
            test_data,
            self.tsv_file,
            write_options=pacsv.WriteOptions(delimiter="\t", include_header=False),
        )

        # Create PSV file
        pacsv.write_csv(
            test_data,
            self.psv_file,
            write_options=pacsv.WriteOptions(delimiter="|", include_header=False),
        )

        # Create unescaped TSV file
        pacsv.write_csv(
            test_data,
            self.unescaped_tsv_file,
            write_options=pacsv.WriteOptions(
                delimiter="\t", include_header=False, quoting_style="none"
            ),
        )

        # Create Parquet file
        papq.write_table(test_data, self.parquet_file)

        # Create Feather file
        paf.write_feather(test_data, self.feather_file)

        # Create JSON file (write as JSONL format)
        df = test_data.to_pandas()
        with open(self.json_file, "w") as f:
            for _, row in df.iterrows():
                json.dump(row.to_dict(), f)
                f.write("\n")

        # Create ORC file
        paorc.write_table(test_data, self.orc_file)

        # Create Avro file
        try:
            import polars as pl

            pl_df = pl.from_arrow(test_data)
            pl_df.write_avro(self.avro_file)
        except ImportError:
            # Skip Avro file creation if polars is not available
            self.avro_file = None

    def _get_filesystems(self, file_path):
        """Get different filesystem implementations for testing."""
        # fsspec AbstractFileSystem
        fsspec_fs = fsspec.filesystem("file")

        # PyArrow filesystem
        import pyarrow.fs as pafs

        pyarrow_fs = pafs.LocalFileSystem()

        # None for automatic inference
        auto_infer_fs = None

        return [
            ("fsspec", fsspec_fs),
            ("pyarrow", pyarrow_fs),
            ("auto_infer", auto_infer_fs),
        ]

    def _assert_table_content(self, table, content_type):
        """Assert that the loaded table has expected content."""
        self.assertEqual(len(table), 5, f"Expected 5 rows for {content_type}")
        self.assertEqual(
            len(table.columns), 4, f"Expected 4 columns for {content_type}"
        )

        # Check column names exist (order might vary for some formats)
        column_names = set(table.column_names)
        expected_columns = {"id", "name", "age", "score"}
        self.assertEqual(
            column_names, expected_columns, f"Column names mismatch for {content_type}"
        )

    def test_csv_all_filesystems(self):
        """Test CSV reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.csv_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.csv_file,
                    ContentType.CSV.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                    column_names=["id", "name", "age", "score"],
                )
                self._assert_table_content(table, f"CSV with {fs_name}")

    def test_tsv_all_filesystems(self):
        """Test TSV reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.tsv_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.tsv_file,
                    ContentType.TSV.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                    column_names=["id", "name", "age", "score"],
                )
                self._assert_table_content(table, f"TSV with {fs_name}")

    def test_psv_all_filesystems(self):
        """Test PSV reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.psv_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.psv_file,
                    ContentType.PSV.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                    column_names=["id", "name", "age", "score"],
                )
                self._assert_table_content(table, f"PSV with {fs_name}")

    def test_unescaped_tsv_all_filesystems(self):
        """Test unescaped TSV reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.unescaped_tsv_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.unescaped_tsv_file,
                    ContentType.UNESCAPED_TSV.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                    column_names=["id", "name", "age", "score"],
                )
                self._assert_table_content(table, f"UNESCAPED_TSV with {fs_name}")

    def test_parquet_all_filesystems(self):
        """Test Parquet reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.parquet_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.parquet_file,
                    ContentType.PARQUET.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                )
                self._assert_table_content(table, f"PARQUET with {fs_name}")

    def test_feather_all_filesystems(self):
        """Test Feather reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.feather_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.feather_file,
                    ContentType.FEATHER.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                )
                self._assert_table_content(table, f"FEATHER with {fs_name}")

    def test_json_all_filesystems(self):
        """Test JSON reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.json_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.json_file,
                    ContentType.JSON.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                )
                self._assert_table_content(table, f"JSON with {fs_name}")

    def test_orc_all_filesystems(self):
        """Test ORC reading with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.orc_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.orc_file,
                    ContentType.ORC.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                )
                self._assert_table_content(table, f"ORC with {fs_name}")

    def test_avro_all_filesystems(self):
        """Test Avro reading with all filesystem types."""
        if self.avro_file is None:
            self.skipTest("Avro file creation skipped (polars not available)")

        for fs_name, filesystem in self._get_filesystems(self.avro_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.avro_file,
                    ContentType.AVRO.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                )
                self._assert_table_content(table, f"AVRO with {fs_name}")

    def test_column_selection_all_filesystems(self):
        """Test column selection works with all filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.parquet_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.parquet_file,
                    ContentType.PARQUET.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                    include_columns=["name", "age"],
                )
                self.assertEqual(
                    len(table.columns), 2, f"Expected 2 columns with {fs_name}"
                )
                self.assertEqual(
                    set(table.column_names),
                    {"name", "age"},
                    f"Column selection failed with {fs_name}",
                )

    def test_kwargs_provider_all_filesystems(self):
        """Test that kwargs providers work with all filesystem types."""

        def schema_provider(content_type, kwargs):
            if content_type == ContentType.CSV.value:
                # Force all columns to be strings
                kwargs["convert_options"] = pacsv.ConvertOptions(
                    column_types={
                        "id": pa.string(),
                        "name": pa.string(),
                        "age": pa.string(),
                        "score": pa.string(),
                    }
                )
            return kwargs

        for fs_name, filesystem in self._get_filesystems(self.csv_file):
            with self.subTest(filesystem=fs_name):
                table = file_to_table(
                    self.csv_file,
                    ContentType.CSV.value,
                    ContentEncoding.IDENTITY.value,
                    filesystem=filesystem,
                    column_names=["id", "name", "age", "score"],
                    pa_read_func_kwargs_provider=schema_provider,
                )
                # Check that all columns are strings
                for field in table.schema:
                    self.assertEqual(
                        field.type,
                        pa.string(),
                        f"Column {field.name} should be string with {fs_name}",
                    )

    def test_filesystem_auto_inference_consistency(self):
        """Test that auto-inferred filesystem produces same results as explicit filesystems."""
        # Use Parquet as it's most reliable across filesystem types

        # Read with auto-inference
        auto_table = file_to_table(
            self.parquet_file,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            filesystem=None,  # Auto-infer
        )

        # Read with explicit fsspec filesystem
        fsspec_fs = fsspec.filesystem("file")
        fsspec_table = file_to_table(
            self.parquet_file,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            filesystem=fsspec_fs,
        )

        # Read with explicit PyArrow filesystem
        import pyarrow.fs as pafs

        pyarrow_fs = pafs.LocalFileSystem()
        pyarrow_table = file_to_table(
            self.parquet_file,
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            filesystem=pyarrow_fs,
        )

        # All should produce equivalent results
        self.assertTrue(
            auto_table.equals(fsspec_table),
            "Auto-inferred result should match fsspec result",
        )
        self.assertTrue(
            auto_table.equals(pyarrow_table),
            "Auto-inferred result should match PyArrow result",
        )

    def test_error_handling_all_filesystems(self):
        """Test error handling works consistently across filesystem types."""
        for fs_name, filesystem in self._get_filesystems(self.parquet_file):
            with self.subTest(filesystem=fs_name):
                # Test unsupported content type
                with self.assertRaises(NotImplementedError):
                    file_to_table(
                        self.parquet_file,
                        "UNSUPPORTED_TYPE",
                        ContentEncoding.IDENTITY.value,
                        filesystem=filesystem,
                    )

                # Test non-existent file
                with self.assertRaises((FileNotFoundError, OSError)):
                    file_to_table(
                        f"{self.tmpdir}/non_existent.parquet",
                        ContentType.PARQUET.value,
                        ContentEncoding.IDENTITY.value,
                        filesystem=filesystem,
                    )



from deltacat.compute.resource_estimation.model import (
    ResourceEstimationMethod,
    EstimatedResources,
    Statistics,
    EstimateResourcesParams,
    OperationType,
)
from deltacat.compute.resource_estimation.manifest import (
    estimate_manifest_entry_column_size_bytes,
    estimate_manifest_entry_num_rows,
    estimate_manifest_entry_size_bytes,
)
from deltacat.compute.resource_estimation.delta import (
    estimate_resources_required_to_process_delta,
)

__all__ = [
    "ResourceEstimationMethod",
    "EstimatedResources",
    "EstimateResourcesParams",
    "Statistics",
    "estimate_resources_required_to_process_delta",
    "estimate_manifest_entry_size_bytes",
    "estimate_manifest_entry_num_rows",
    "estimate_manifest_entry_column_size_bytes",
    "OperationType",
]



from __future__ import annotations
from typing import Dict, List, Any, Optional
from deltacat.compute.converter.model.convert_input_files import ConvertInputFiles
from fsspec import AbstractFileSystem


class ConvertInput(Dict):
    @staticmethod
    def of(
        convert_input_files: ConvertInputFiles,
        convert_task_index: int,
        iceberg_table_warehouse_prefix: str,
        identifier_fields: List[str],
        table_io: Any,
        table_metadata: Any,
        compact_previous_position_delete_files: bool,
        enforce_primary_key_uniqueness: bool,
        position_delete_for_multiple_data_files: bool,
        max_parallel_data_file_download: int,
        filesystem: Optional[AbstractFileSystem],
        s3_client_kwargs: Optional[Dict[str, Any]],
        task_memory: float,
    ) -> ConvertInput:

        result = ConvertInput()
        result["convert_input_files"] = convert_input_files
        result["convert_task_index"] = convert_task_index
        result["identifier_fields"] = identifier_fields
        result["iceberg_table_warehouse_prefix"] = iceberg_table_warehouse_prefix
        result["table_io"] = table_io
        result["table_metadata"] = table_metadata
        result[
            "compact_previous_position_delete_files"
        ] = compact_previous_position_delete_files
        result["enforce_primary_key_uniqueness"] = enforce_primary_key_uniqueness
        result[
            "position_delete_for_multiple_data_files"
        ] = position_delete_for_multiple_data_files
        result["max_parallel_data_file_download"] = max_parallel_data_file_download
        result["filesystem"] = filesystem
        result["s3_client_kwargs"] = s3_client_kwargs
        result["task_memory"] = task_memory

        return result

    @property
    def convert_input_files(self) -> ConvertInputFiles:
        return self["convert_input_files"]

    @property
    def identifier_fields(self) -> List[str]:
        return self["identifier_fields"]

    @property
    def convert_task_index(self) -> int:
        return self["convert_task_index"]

    @property
    def iceberg_table_warehouse_prefix(self) -> str:
        return self["iceberg_table_warehouse_prefix"]

    @property
    def table_io(self) -> Any:
        return self["table_io"]

    @property
    def table_metadata(self) -> Any:
        return self["table_metadata"]

    @property
    def compact_previous_position_delete_files(self) -> bool:
        return self["compact_previous_position_delete_files"]

    @property
    def enforce_primary_key_uniqueness(self) -> bool:
        return self["enforce_primary_key_uniqueness"]

    @property
    def position_delete_for_multiple_data_files(self) -> bool:
        return self["position_delete_for_multiple_data_files"]

    @property
    def max_parallel_data_file_download(self) -> int:
        return self["max_parallel_data_file_download"]

    @property
    def filesystem(self) -> Optional[AbstractFileSystem]:
        return self["filesystem"]

    @property
    def s3_client_kwargs(self) -> Optional[Dict[str, Any]]:
        return self["s3_client_kwargs"]

    @property
    def task_memory(self) -> float:
        return self["task_memory"]



import pytest
from pyspark.sql import SparkSession
import os
import ray
from pyiceberg.catalog import Catalog, load_catalog


@pytest.fixture
def spark():
    import importlib.metadata

    spark_version = ".".join(importlib.metadata.version("pyspark").split(".")[:2])
    scala_version = "2.12"
    iceberg_version = "1.6.0"

    os.environ["PYSPARK_SUBMIT_ARGS"] = (
        f"--packages org.apache.iceberg:iceberg-spark-runtime-{spark_version}_{scala_version}:{iceberg_version},"
        f"org.apache.iceberg:iceberg-aws-bundle:{iceberg_version} pyspark-shell"
    )
    os.environ["AWS_REGION"] = "us-east-1"
    os.environ["AWS_ACCESS_KEY_ID"] = "admin"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "password"

    spark = (
        SparkSession.builder.appName("PyIceberg integration test")
        .config("spark.sql.session.timeZone", "UTC")
        .config(
            "spark.sql.extensions",
            "org.apache.iceberg.spark.extensions.IcebergSparkSessionExtensions",
        )
        .config(
            "spark.sql.catalog.integration", "org.apache.iceberg.spark.SparkCatalog"
        )
        .config(
            "spark.sql.catalog.integration.catalog-impl",
            "org.apache.iceberg.rest.RESTCatalog",
        )
        .config("spark.sql.catalog.integration.cache-enabled", "false")
        .config("spark.sql.catalog.integration.uri", "http://localhost:8181")
        .config(
            "spark.sql.catalog.integration.io-impl",
            "org.apache.iceberg.aws.s3.S3FileIO",
        )
        .config("spark.sql.catalog.integration.warehouse", "s3://warehouse/wh/")
        .config("spark.sql.catalog.integration.s3.endpoint", "http://localhost:9000")
        .config("spark.sql.catalog.integration.s3.path-style-access", "true")
        .config("spark.sql.defaultCatalog", "integration")
        .config("spark.sql.catalog.hive", "org.apache.iceberg.spark.SparkCatalog")
        .config("spark.sql.catalog.hive.type", "hive")
        .config("spark.sql.catalog.hive.uri", "http://localhost:9083")
        .config("spark.sql.catalog.hive.io-impl", "org.apache.iceberg.aws.s3.S3FileIO")
        .config("spark.sql.catalog.hive.warehouse", "s3://warehouse/hive/")
        .config("spark.sql.catalog.hive.s3.endpoint", "http://localhost:9000")
        .config("spark.sql.catalog.hive.s3.path-style-access", "true")
        .config("spark.sql.execution.arrow.pyspark.enabled", "true")
        .getOrCreate()
    )

    return spark


@pytest.fixture(scope="session")
def session_catalog() -> Catalog:
    return load_catalog(
        "local",
        **{
            "type": "rest",
            "uri": "http://localhost:8181",
            "s3.endpoint": "http://localhost:9000",
            "s3.access-key-id": "admin",
            "s3.secret-access-key": "password",
        },
    )


@pytest.fixture(autouse=True, scope="module")
def setup_ray_cluster():
    ray.init(local_mode=True, ignore_reinit_error=True)
    yield
    ray.shutdown()



import pyarrow.compute as pc

import deltacat.compute.converter.utils.iceberg_columns as sc
import pyarrow as pa

from collections import defaultdict
import ray
import logging
from deltacat.compute.converter.model.convert_input import ConvertInput
from deltacat.compute.converter.steps.dedupe import dedupe_data_files
from deltacat.compute.converter.utils.io import write_sliced_table
from deltacat.compute.converter.utils.io import (
    download_data_table_and_append_iceberg_columns,
)
from deltacat.compute.converter.utils.converter_session_utils import (
    partition_value_record_to_partition_value_string,
    sort_data_files_maintaining_order,
)
from deltacat.compute.converter.pyiceberg.overrides import (
    parquet_files_dict_to_iceberg_data_files,
)
from deltacat.compute.converter.model.convert_result import ConvertResult
from pyiceberg.manifest import DataFileContent
from deltacat import logs
from fsspec import AbstractFileSystem
from typing import List, Dict, Tuple, Optional, Any
from deltacat.utils.resources import get_current_process_peak_memory_usage_in_bytes
from deltacat.compute.converter.model.convert_input_files import (
    DataFileList,
    DataFileListGroup,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote
def convert(convert_input: ConvertInput) -> ConvertResult:
    convert_input_files = convert_input.convert_input_files
    convert_task_index = convert_input.convert_task_index
    iceberg_table_warehouse_prefix = convert_input.iceberg_table_warehouse_prefix
    identifier_fields = convert_input.identifier_fields
    table_io = convert_input.table_io
    table_metadata = convert_input.table_metadata
    compact_previous_position_delete_files = (
        convert_input.compact_previous_position_delete_files
    )
    position_delete_for_multiple_data_files = (
        convert_input.position_delete_for_multiple_data_files
    )
    max_parallel_data_file_download = convert_input.max_parallel_data_file_download
    filesystem = convert_input.filesystem
    s3_client_kwargs = convert_input.s3_client_kwargs
    task_memory = convert_input.task_memory

    if not position_delete_for_multiple_data_files:
        raise NotImplementedError(
            f"Distributed file level position delete compute is not supported yet"
        )
    if compact_previous_position_delete_files:
        raise NotImplementedError(f"Compact previous position delete not supported yet")

    logger.info(f"Starting convert task index: {convert_task_index}")

    applicable_data_files = convert_input_files.applicable_data_files
    applicable_equality_delete_files = (
        convert_input_files.applicable_equality_delete_files
    )

    all_data_files_for_this_bucket = convert_input_files.all_data_files_for_dedupe

    partition_value_str = partition_value_record_to_partition_value_string(
        convert_input_files.partition_value
    )
    partition_value = convert_input_files.partition_value

    if partition_value_str:
        iceberg_table_warehouse_prefix_with_partition = (
            f"{iceberg_table_warehouse_prefix}/{partition_value_str}"
        )
    else:
        iceberg_table_warehouse_prefix_with_partition = (
            f"{iceberg_table_warehouse_prefix}"
        )

    enforce_primary_key_uniqueness = convert_input.enforce_primary_key_uniqueness
    total_pos_delete_table = []
    data_table_after_converting_equality_delete = []
    if applicable_equality_delete_files:
        (
            pos_delete_after_converting_equality_delete,
            data_table_after_converting_equality_delete,
        ) = compute_pos_delete_with_limited_parallelism(
            data_files_list=applicable_data_files,
            identifier_columns=identifier_fields,
            equality_delete_files_list=applicable_equality_delete_files,
            iceberg_table_warehouse_prefix_with_partition=iceberg_table_warehouse_prefix_with_partition,
            convert_task_index=convert_task_index,
            max_parallel_data_file_download=max_parallel_data_file_download,
            s3_file_system=filesystem,
            s3_client_kwargs=s3_client_kwargs,
        )
        if pos_delete_after_converting_equality_delete:
            total_pos_delete_table.append(pos_delete_after_converting_equality_delete)

    if enforce_primary_key_uniqueness:
        data_files_downloaded_during_convert = []
        if applicable_data_files:
            for file_list in applicable_data_files:
                for file in file_list:
                    data_files_downloaded_during_convert.append(file)

        data_files_to_dedupe = get_additional_applicable_data_files(
            all_data_files=all_data_files_for_this_bucket,
            data_files_downloaded=data_files_downloaded_during_convert,
        )

        dedupe_file_size_bytes = sum(
            data_file.file_size_in_bytes for _, data_file in data_files_to_dedupe
        )
        logger.info(
            f"Total on-disk size of files to dedupe: {dedupe_file_size_bytes} bytes"
        )

        logger.info(
            f"[Convert task {convert_task_index}]: Got {len(data_files_to_dedupe)} files to dedupe."
        )

        (
            pos_delete_after_dedupe,
            data_file_to_dedupe_record_count,
            data_file_to_dedupe_size,
        ) = dedupe_data_files(
            data_file_to_dedupe=data_files_to_dedupe,
            identifier_columns=identifier_fields,
            remaining_data_table_after_convert=data_table_after_converting_equality_delete,
            merge_sort_column=sc._ORDERED_RECORD_IDX_COLUMN_NAME,
            s3_client_kwargs=s3_client_kwargs,
        )
        logger.info(
            f"[Convert task {convert_task_index}]: Dedupe produced {len(pos_delete_after_dedupe)} position delete records."
        )
        total_pos_delete_table.append(pos_delete_after_dedupe)

    total_pos_delete = pa.concat_tables(total_pos_delete_table)

    logger.info(
        f"[Convert task {convert_task_index}]: Total position delete produced:{len(total_pos_delete)}"
    )

    to_be_added_files_list = []
    if total_pos_delete:
        to_be_added_files_list_parquet = write_sliced_table(
            table=total_pos_delete,
            base_path=iceberg_table_warehouse_prefix_with_partition,
            table_writer_kwargs={},
            filesystem=filesystem,
        )

        to_be_added_files_dict = defaultdict()
        to_be_added_files_dict[partition_value] = to_be_added_files_list_parquet

        logger.info(
            f"[Convert task {convert_task_index}]: Produced {len(to_be_added_files_list_parquet)} position delete files."
        )
        file_content_type = DataFileContent.POSITION_DELETES
        to_be_added_files_list = parquet_files_dict_to_iceberg_data_files(
            io=table_io,
            table_metadata=table_metadata,
            files_dict=to_be_added_files_dict,
            file_content_type=file_content_type,
        )

    to_be_delete_files_dict = defaultdict()

    if applicable_equality_delete_files:
        to_be_delete_files_dict[partition_value] = [
            equality_delete_file[1]
            for equality_delete_list in applicable_equality_delete_files
            for equality_delete_file in equality_delete_list
        ]

    if not enforce_primary_key_uniqueness:
        data_file_to_dedupe_record_count = 0
        data_file_to_dedupe_size = 0

    peak_memory_usage_bytes = (
        get_current_process_peak_memory_usage_in_bytes()
    )  # Convert KB to bytes
    memory_usage_percentage = (peak_memory_usage_bytes / task_memory) * 100

    logger.info(
        f"[Convert task {convert_task_index}]: Memory usage stats - "
        f"Peak memory usage: {peak_memory_usage_bytes} bytes, "
        f"Allocated task memory: {convert_input.task_memory} bytes, "
        f"Usage percentage: {memory_usage_percentage:.2f}%"
    )

    convert_res = ConvertResult.of(
        convert_task_index=convert_task_index,
        to_be_added_files=to_be_added_files_list,
        to_be_deleted_files=to_be_delete_files_dict,
        position_delete_record_count=len(total_pos_delete),
        input_data_files_record_count=data_file_to_dedupe_record_count,
        input_data_files_hash_columns_in_memory_sizes=data_file_to_dedupe_size,
        position_delete_in_memory_sizes=int(total_pos_delete.nbytes),
        position_delete_on_disk_sizes=sum(
            file.file_size_in_bytes for file in to_be_added_files_list
        ),
        input_data_files_on_disk_size=dedupe_file_size_bytes,
        peak_memory_usage_bytes=peak_memory_usage_bytes,
        memory_usage_percentage=memory_usage_percentage,
    )
    return convert_res


def get_additional_applicable_data_files(
    all_data_files: DataFileList,
    data_files_downloaded: DataFileList,
) -> DataFileList:
    data_file_to_dedupe = []
    assert len(set(all_data_files)) >= len(set(data_files_downloaded)), (
        f"Length of all data files ({len(set(all_data_files))}) should never be less than "
        f"the length of candidate equality delete data files ({len(set(data_files_downloaded))})"
    )
    if data_files_downloaded:
        # set1.difference(set2) returns elements in set1 but not in set2
        data_file_to_dedupe.extend(
            list(set(data_file_to_dedupe).difference(set(data_files_downloaded)))
        )
    else:
        data_file_to_dedupe = all_data_files
    return data_file_to_dedupe


def filter_rows_to_be_deleted(
    equality_delete_table: Optional[pa.Table],
    data_file_table: Optional[pa.Table],
    identifier_columns: List[str],
) -> Tuple[Optional[pa.Table], Optional[pa.Table]]:
    identifier_column = sc._IDENTIFIER_COLUMNS_HASH_COLUMN_NAME
    if equality_delete_table and data_file_table:
        equality_deletes = pc.is_in(
            data_file_table[identifier_column],
            equality_delete_table[identifier_column],
        )
        data_file_record_remaining = pc.invert(
            pc.is_in(
                data_file_table[identifier_column],
                equality_delete_table[identifier_column],
            )
        )
        position_delete_table = data_file_table.filter(equality_deletes)
        remaining_data_table = data_file_table.filter(data_file_record_remaining)

        position_delete_table = position_delete_table.drop(
            [sc._IDENTIFIER_COLUMNS_HASH_COLUMN_NAME]
        )
        assert len(position_delete_table) + len(remaining_data_table) == len(
            data_file_table
        ), (
            f"Expected undeleted data file record count plus length of pos deletes to match original data file record count of {len(data_file_table)}, "
            f"but found {len(position_delete_table)} pos deletes + {len(remaining_data_table)} equality deletes."
        )

    return position_delete_table, remaining_data_table


def compute_pos_delete_converting_equality_deletes(
    equality_delete_table: Optional[pa.Table],
    data_file_table: Optional[pa.Table],
    identifier_columns: List[str],
    iceberg_table_warehouse_prefix_with_partition: str,
    s3_file_system: Optional[AbstractFileSystem],
) -> Tuple[Optional[pa.Table], Optional[pa.Table]]:
    new_position_delete_table, remaining_data_table = filter_rows_to_be_deleted(
        data_file_table=data_file_table,
        equality_delete_table=equality_delete_table,
        identifier_columns=identifier_columns,
    )
    if new_position_delete_table:
        logger.info(
            f"Length of position delete table after converting from equality deletes:{len(new_position_delete_table)}"
        )
        return new_position_delete_table, remaining_data_table
    elif not remaining_data_table:
        return None, None
    else:
        return None, remaining_data_table


def compute_pos_delete_with_limited_parallelism(
    data_files_list: DataFileListGroup,
    identifier_columns: List[str],
    equality_delete_files_list: DataFileListGroup,
    iceberg_table_warehouse_prefix_with_partition: str,
    convert_task_index: int,
    max_parallel_data_file_download: int,
    s3_file_system: Optional[AbstractFileSystem],
    s3_client_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Optional[pa.Table], Optional[pa.Table]]:
    assert len(data_files_list) == len(equality_delete_files_list), (
        f"Number of lists of data files should equal to number of list of equality delete files, "
        f"But got {len(data_files_list)} data files lists vs {len(equality_delete_files_list)}."
    )

    new_pos_delete_table_total = []
    for data_files, equality_delete_files in zip(
        data_files_list, equality_delete_files_list
    ):
        data_table_total = []

        # Sort data files by file sequence number first, then file path to
        # make sure files having same sequence number are deterministically sorted
        data_files = sort_data_files_maintaining_order(data_files=data_files)

        for data_file in data_files:
            data_table = download_data_table_and_append_iceberg_columns(
                file=data_file[1],
                columns_to_download=identifier_columns,
                additional_columns_to_append=[
                    sc._FILE_PATH_COLUMN_NAME,
                    sc._ORDERED_RECORD_IDX_COLUMN_NAME,
                ],
                s3_client_kwargs=s3_client_kwargs,
            )
            data_table_total.append(data_table)
        data_table_total = pa.concat_tables(data_table_total)

        equality_delete_table_total = []
        for equality_delete in equality_delete_files:
            equality_delete_table = download_data_table_and_append_iceberg_columns(
                file=equality_delete[1],
                columns_to_download=identifier_columns,
                s3_client_kwargs=s3_client_kwargs,
            )
            equality_delete_table_total.append(equality_delete_table)
        equality_delete_table_total = pa.concat_tables(equality_delete_table_total)

        (
            new_pos_delete_table,
            remaining_data_table,
        ) = compute_pos_delete_converting_equality_deletes(
            equality_delete_table=equality_delete_table_total,
            data_file_table=data_table_total,
            iceberg_table_warehouse_prefix_with_partition=iceberg_table_warehouse_prefix_with_partition,
            identifier_columns=identifier_columns,
            s3_file_system=s3_file_system,
        )
        new_pos_delete_table_total.append(new_pos_delete_table)

    if new_pos_delete_table_total:
        new_pos_delete_table_total = pa.concat_tables(new_pos_delete_table_total)

    logger.info(
        f"[Convert task {convert_task_index}]: Find deletes got {len(data_table_total)} data table records, "
        f"{len(equality_delete_table_total)} equality deletes as input, "
        f"Produced {len(new_pos_delete_table_total)} position deletes based off find deletes input."
    )

    if not new_pos_delete_table_total:
        logger.info("No records deleted based on equality delete convertion")

    if not remaining_data_table:
        logger.info("No data table remaining after converting equality deletes")

    return new_pos_delete_table_total, remaining_data_table
